From 4c920861f07e0f4e6e21e24e5f6a9060054c3232 Mon Sep 17 00:00:00 2001
From: Michael Paquier <michael@paquier.xyz>
Date: Tue, 15 Jul 2025 12:01:01 +0900
Subject: [PATCH v4 1/3] Fix infinite wait when reading partially-written WAL
 record

If a crash occurs while writing a WAL record that spans multiple pages,
the recovery process marks the page with the
XLP_FIRST_IS_OVERWRITE_CONTRECORD flag.  However, logical decoding
currently attempts to read the full WAL record based on its expected
size before checking this flag, which can lead to an infinite wait if
the remaining data is never written (e.g., no activity after crash).

This patch updates the logic to first read the page header and check for
the XLP_FIRST_IS_OVERWRITE_CONTRECORD flag before attempting to
reconstruct the full WAL record.  If the flag is set, decoding correctly
identifies the record as incomplete and avoids waiting for WAL data that
will never arrive.
---
 src/backend/access/transam/xlogreader.c | 20 +++++++++++++++-----
 1 file changed, 15 insertions(+), 5 deletions(-)

diff --git a/src/backend/access/transam/xlogreader.c b/src/backend/access/transam/xlogreader.c
index ac1f801b1eb0..2e5361ff4fdd 100644
--- a/src/backend/access/transam/xlogreader.c
+++ b/src/backend/access/transam/xlogreader.c
@@ -723,11 +723,12 @@ restart:
 			/* Calculate pointer to beginning of next page */
 			targetPagePtr += XLOG_BLCKSZ;
 
-			/* Wait for the next page to become available */
-			readOff = ReadPageInternal(state, targetPagePtr,
-									   Min(total_len - gotlen + SizeOfXLogShortPHD,
-										   XLOG_BLCKSZ));
-
+			/*
+			 * Read the page header before processing the record data, to be
+			 * able to handle the case where the previous record ended as
+			 * being a partial one.
+			 */
+			readOff = ReadPageInternal(state, targetPagePtr, SizeOfXLogShortPHD);
 			if (readOff == XLREAD_WOULDBLOCK)
 				return XLREAD_WOULDBLOCK;
 			else if (readOff < 0)
@@ -776,6 +777,15 @@ restart:
 				goto err;
 			}
 
+			/* Wait for the next page to become available */
+			readOff = ReadPageInternal(state, targetPagePtr,
+									   Min(total_len - gotlen + SizeOfXLogShortPHD,
+										   XLOG_BLCKSZ));
+			if (readOff == XLREAD_WOULDBLOCK)
+				return XLREAD_WOULDBLOCK;
+			else if (readOff < 0)
+				goto err;
+
 			/* Append the continuation from this page to the buffer */
 			pageHeaderSize = XLogPageHeaderSize(pageHeader);
 
-- 
2.50.0

